/*
 *  Created on: July 12, 2021
 *  Author: KmolYuan
 */

#ifndef DERIVATIVES_H
#define DERIVATIVES_H

#include "calc.h"

void derivatives(double **, double *, size_t, Constraint *, size_t);

#endif  // DERIVATIVES_H
